from email import encoders
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from notify_mi.providers import PROVIDERS
from pathlib import Path
import smtplib, ssl
from notify_mi import helper

''' 
This program uses gmail to send a SMS/MMS message to a phone number
with a known service provider and optionally to an email address.
'''

def send_message(
    message,
    sender_credentials,
    phone_number = None,
    phone_provider = None, 
    subject = None, 
    send_to = None, 
    threaded = False, 
    file_attachment = None):
    
    '''
    Send a message using gmail via SMS/MMS, email, or both
    
    :param            str : message : message to be sent
    :param          tuple : sender_credentials : gmail credentials
    :param <optional> str : phone_number : phone number of reciever
    :param <optional> str : phone_provider : cell provider of reciever number
    :param <optional> str : subject : add a subject to message
    :param <optional> str : file_attachment : path to file to attach to message
    :param <optional> bool: threaded : user can decide if they want to not block main thread
    :param <optional> str : send_to : email address to send to
    '''
    
    # allow user to not block the main thread when calling
    if threaded:
        helper.threaded(__send_message_via_email(
            message, sender_credentials, phone_number, phone_provider, 
            subject, send_to, file_attachment))
    else:
        __send_message_via_email(
            message, sender_credentials, phone_number, phone_provider, 
            subject, send_to, file_attachment)


def __send_message_via_email(
    message,
    sender_credentials,
    phone_number = None,
    phone_provider = None, 
    subject = None, 
    send_to = None, 
    file_attachment = None):
    
    # ensure that credentials and info added meets specifications
    __check_for_exceptions(phone_number, phone_provider, send_to)
    
    # initialize variables needed
    _phone_number: str = phone_number
    _message: str = message
    file_attachment: str = file_attachment
    _subject: str = subject
    _phone_provider: str = phone_provider
    _sender_credentials: tuple = sender_credentials
    smtp_server: str = helper.SMTP_GMAIL
    smtp_port: int = helper.SMTP_PORT
    if file_attachment is not None:
        attachment = Path(file_attachment)
        mime_subtype: str = attachment.suffix
        mime_maintype: str = helper.find_ext_mime_type(mime_subtype)
        file_name: str = attachment.name
    
    # determine type(s) of messages being sent
    is_sending_text_message = True if None not in [_phone_number, _phone_provider] else False
    is_sending_email = True if send_to is not None and send_to is not helper.EMPTY else False
        
    # retrieve information needed to send message
    sender_email, email_password = _sender_credentials
    # get message type (sms/mms) based on provider
    # some do not have mms capabilities
    if is_sending_text_message:
        message_type = helper.MESSAGE_TYPE[0] \
            if PROVIDERS.get(_phone_provider).get(helper.MMS_SUPPORT_KEY) \
            else helper.MESSAGE_TYPE[1]
        # create receiver email based on their phone number and carrier
        receiver_phone_number = f'{_phone_number}@{PROVIDERS.get(_phone_provider).get(message_type)}'
    
    # create gmail body
    email_message = MIMEMultipart()
    if _subject is not None:
        email_message["Subject"] = _subject
    email_message["From"] = sender_email
    email_message.attach(MIMEText(_message, helper.TEXT_TYPE))
    
    # if file was included, attach to message
    if file_attachment is not None and Path(file_attachment).is_file:
        # gmail does not allow a file > 1 MB to be attached
        # therefore the entire message will not be sent
        if attachment.stat().st_size > 1 * helper.MB:
            raise helper.FileSizeExceeded
        
        # open file being sent and attach to email_message
        with open(file_attachment, helper.READ_BINARY) as attachment:
            part = MIMEBase(mime_maintype, mime_subtype)
            part.set_payload(attachment.read())
            encoders.encode_base64(part)
            part.add_header("Content-Disposition", f"attachment; filename={file_name}",)
            email_message.attach(part)

    # send the message
    with smtplib.SMTP_SSL(smtp_server, smtp_port, context = ssl.create_default_context()) as email:
        # login to gmail
        email.login(sender_email, email_password)
        
        # send SMS/MMS message
        if is_sending_text_message:
            # send email with body and attachment
            email.sendmail(sender_email, receiver_phone_number, email_message.as_string())
            
        # send to email address
        if is_sending_email:
            # send email with body and attachment
            email.sendmail(sender_email, send_to, email_message.as_string())


def __check_for_exceptions(phone_number, phone_provider, send_to): 
    # check if at least one message type is being sent
    # check if all paramters is empty
    if None in [phone_number, phone_provider]:
        # if either phone number or phone provider is not empty
        # specify that one of them is missing
        if phone_number is not None or phone_provider is not None:
            raise helper.MessageTextError
        # else, all of them is empty, thus give user a specific error message
        elif send_to is None:
            raise helper.MessageError
    elif phone_number is not None and phone_provider is not None:
        # ensure reciever email is formatted correctly
        if send_to is not None and send_to is not helper.EMPTY:
            if helper.AT_SYMBOL not in send_to or helper.EMAIL_SUFFIX not in send_to:
                raise helper.EmailFormatError
        # verify phone number is formatted correctly
        if len(phone_number) != 10 or not phone_number.isdigit():
            raise helper.PhoneNumberError
        # verify provider given is found in providers.py
        if PROVIDERS.get(phone_provider) is None:
            raise helper.ProviderNotRecognized
        


# List of Phone Providers
# Select From: 
# "AT&T", "Boost Mobile", "C-Spire", "Cricket Wireless", 
# "Consumer Cellular", "Google Project Fi", "Metro PCS", 
# "Mint Mobile", "Page Plus", "Republic Wireless", "Sprint",
# "Straight Talk", "T-Mobile", "Ting", "Tracfone", "U.S. 
#  Cellular", "Verizon", "Virgin Mobile", and "Xfinity Mobile"

PROVIDERS = {
    "AT&T": {"sms": "txt.att.net", "mms": "mms.att.net", "mms_support": True},
    "Boost Mobile": {
        "sms": "sms.myboostmobile.com",
        "mms": "myboostmobile.com",
        "mms_support": True,
    },
    "C-Spire": {"sms": "cspire1.com", "mms_support": False},
    "Cricket Wireless": {
        "sms": "sms.cricketwireless.net ",
        "mms": "mms.cricketwireless.net",
        "mms_support": True,
    },
    "Consumer Cellular": {"sms": "mailmymobile.net", "mms_support": False},
    "Google Project Fi": {"sms": "msg.fi.google.com", "mms_support": False},
    "Metro PCS": {"sms": "mymetropcs.com", "mms_support": True},
    "Mint Mobile": {"sms": "mailmymobile.net", "mms_support": False},
    "Page Plus": {
        "sms": "vtext.com",
        "mms": "mypixmessages.com",
        "mms_support": True,
    },
    "Republic Wireless": {
        "sms": "text.republicwireless.com",
        "mms_support": False,
    },
    "Sprint": {
        "sms": "messaging.sprintpcs.com",
        "mms": "pm.sprint.com",
        "mms_support": True,
    },
    "Straight Talk": {
        "sms": "vtext.com",
        "mms": "mypixmessages.com",
        "mms_support": True,
    },
    "T-Mobile": {"sms": "tmomail.net", "mms_support": False},
    "Ting": {"sms": "message.ting.com", "mms_support": False},
    "Tracfone": {"sms": "", "mms": "mmst5.tracfone.com", "mms_support": True},
    "U.S. Cellular": {
        "sms": "email.uscc.net",
        "mms": "mms.uscc.net",
        "mms_support": True,
    },
    "Verizon": {"sms": "vtext.com", "mms": "vzwpix.com", "mms_support": True},
    "Virgin Mobile": {
        "sms": "vmobl.com",
        "mms": "vmpix.com",
        "mms_support": True,
    },
    "Xfinity Mobile": {
        "sms": "vtext.com",
        "mms": "mypixmessages.com",
        "mms_support": True,
    },
}



import threading
import functools

'''
This is a helper class that contains constants 
and methods used throughout multiple files.
'''

# constants
EMPTY = ""
SMTP_GMAIL = "smtp.gmail.com"
SMTP_PORT = 465
MESSAGE_TYPE = ["mms", "sms"]
MMS_SUPPORT_KEY = "mms_support"
TEXT_TYPE = "plain"
READ_BINARY = 'rb'
AT_SYMBOL = "@"
EMAIL_SUFFIX = ".com"
MB = 1000000

# exceptions
class FileSizeExceeded(Exception):
    def __str__(self):
        return "Attachment cannot be more than 1MB."

class PhoneNumberError(Exception):
    def __str__(self):
        return "Phone number should have 10 numbers only."
    
class ProviderNotRecognized(Exception):
    def __str__(self):
        return "Select a phone provider from providers.py file."

class ExtensionNotFound(Exception):
    def __str__(self):
        return "Attached file has an extension that is not recognized."
    
class EmailFormatError(Exception):
    def __str__(self):
        return "The receiver email should follow format -> name@email.com."

class MessageError(Exception):
    def __str__(self):
        return "Message type not specified. Pick text, email, or both and include the respective paramters."

class MessageTextError(Exception):
    def __str__(self):
        return "Phone number AND phone provider required for sending a text message."

class MessageEmailError(Exception):
    def __str__(self):
        return "Receiver's email is required to send a message."

# decorator to automatically launch a function in a thread
def threaded(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        thread = threading.Thread(target=func, args=args, kwargs=kwargs)
        thread.start()
        return thread
    return wrapper

extensions_and_mime_types = [
    (".aac", "audio/aac"),    
    (".abw", "application/x-abiword"),    
    (".arc", "application/x-freearc"),    
    (".avif", "image/avif"),    
    (".avi", "video/x-msvideo"),    
    (".azw", "application/vnd.amazon.ebook"),    
    (".bin", "application/octet-stream"),    
    (".bmp", "image/bmp"),    
    (".bz", "application/x-bzip"),    
    (".bz2", "application/x-bzip2"),    
    (".cda", "application/x-cdf"),    
    (".csh", "application/x-csh"),    
    (".css", "text/css"),    
    (".csv", "text/csv"),    
    (".doc", "application/msword"),    
    (".docx", "application/vnd.openxmlformats-"),
    (".eot", "application/vnd.ms-fontobject"),    
    (".epub", "application/epub+zip"),    
    (".gz", "application/gzip"),    
    (".gif", "image/gif"),    
    (".htm, .html", "text/html"),    
    (".ico", "image/vnd.microsoft.icon"),    
    (".ics", "text/calendar"),    
    (".jar", "application/java-archive"),    
    (".jpeg, .jpg", "image/jpeg"),    
    (".js", "text/javascript"),    
    (".json", "application/json"),    
    (".jsonld", "application/ld+json"),    
    (".mid, .midi", "audio/midi, audio/x-midi"),    
    (".mjs", "text/javascript"),    
    (".mp3", "audio/mpeg"),    
    (".mp4", "video/mp4"),    
    (".mpeg", "video/mpeg"),
    (".mpkg", "application/vnd.apple.installer+xml"),    
    (".odp", "application/vnd.oasis.opendocument.presentation"),    
    (".ods", "application/vnd.oasis.opendocument.spreadsheet"),    
    (".odt", "application/vnd.oasis.opendocument.text"),    
    (".oga", "audio/ogg"),    (".ogv", "video/ogg"),    
    (".ogx", "application/ogg"),    (".opus", "audio/opus"),   
    (".otf", "font/otf"),    
    (".png", "image/png"),    
    (".pdf", "application/pdf"),   
    (".php", "application/x-httpd-php"),    
    (".ppt", "application/vnd.ms-powerpoint"),    
    (".pptx", "application/vnd.openxmlformats-officedocument.presentationml.presentation"),   
    (".rar", "application/vnd.rar"), 
    (".rtf", "application/rtf"),    
    (".sh", "application/x-sh"),
    (".svg", "image/svg+xml"),    
    (".tar", "application/x-tar"),    
    (".tif, .tiff", "image/tiff"),    
    (".ts", "video/mp2t"),
    (".ttf", "font/ttf"),    
    (".txt", "text/plain"),    
    (".vsd", "application/vnd.visio"),    
    (".wav", "audio/wav"),    
    (".weba", "audio/webm"),    
    (".webm", "video/webm"),    
    (".webp", "image/webp"),    
    (".woff", "font/woff"),    
    (".woff2", "font/woff2"),    
    (".xhtml", "application/xhtml+xml"),    
    (".xls", "application/vnd.ms-excel"),    
    (".xlsx", "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"),    
    (".xml", "application/xml"),    
    (".xul", "application/vnd.mozilla.xul+xml"),    
    (".zip", "application/zip"),    
    (".3gp", "video/3gpp"),    
    (".3g2", "video/3gpp2"),    
    (".7z", "application/x-7z-compressed")]

# param: find_ext -> desired extension to find 
# return: mime type for the desired extension
def find_ext_mime_type(find_ext):
    for extension_and_mime_type in extensions_and_mime_types:
        extension = extension_and_mime_type[0]
        if find_ext == extension:
            return extension_and_mime_type[1].split("/")[0]
    raise ExtensionNotFound
